/*
 * TheoraDecoder
 *
 * Copyright (C) 2008 Joern Seger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "theoraDecoder.h"

#ifdef HAVE_LIBTHEORA

#include <sstream>
#include <iostream>

#include <theora/theora.h>

#include "mediaOutputDecoder.h"
#include "oggBOSExtractorFactory.h"

TheoraDecoder::TheoraDecoder(int8 _streamID)
 : MediaOutputDecoder(_streamID), initCount(0)
{
//  init();
}

TheoraDecoder::~TheoraDecoder()
{
}

void TheoraDecoder::initDecoder(StreamConfig& config)
{

  /* initialize the info and comment handler structs */
  theora_info_init(&theoraInfo);
  theora_comment_init(&theoraComment);

  /* initialize the packet counter */
  packetCount = 0;

  /* Konfiguration des Decoders */
  for (uint8 i(0); i<config.headerList.size(); ++i) {

    /* Einfügen der Header
     * Fehlermeldung, wenn die Daten nicht zum aktuellen Codec passen */
    if (theora_decode_header(&theoraInfo, &theoraComment, config.headerList[i].obj()) != 0) {

      std::cerr << "TheoraDecoder::operator<<: packet is not a header\n";

      /* initialization is aborted */
      return; // throw ...
    }

  }

  /* finish initialization */
  theora_decode_init(&theoraState, &theoraInfo);

  /* set the state machine */
  setConfigured();

}

MediaOutputDecoder& TheoraDecoder::operator<<(OggPacket packet)
{
  /* if the stream is not initialized, initialize the first structs */
  if (!isConfigured()) {
    std::cerr << "TheoraDecoder::operator<<: Cannot handle packet, decoder not initialized\n";
    return(*this);
  }

  /* while inserting data into the stream, we do not
   * decode. We just store the packets and will decode them
   * on demand */
  packetList.push_back(packet);

  /* has there not been a packet in the queue before */
  if (isEmpty()) {

    /* set the internal state */
    setAvailable();

  }

  /* count the video packets, to have a gimps of the actual position */
  packetCount++;
  return(*this);
}

bool TheoraDecoder::isNextPacketKeyframe()
{
  return((theora_packet_iskeyframe(packetList.front().obj()) == 1));
}

bool TheoraDecoder::isPacketKeyframe(OggPacket packet)
{
  return((theora_packet_iskeyframe(packet.obj()) == 1));
}

uint32 TheoraDecoder::getPositionOfNextPacket()
{
  if (isEmpty())
    return(0xFFFFFFFF);

  return(packetCount - packetList.size());
}

TheoraDecoder& TheoraDecoder::operator>>(TheoraVideoPacket& picture)
{
  if (!isConfigured()) {
    std::cerr << "TheoraDecoder::operator>>: Theora Stream is not initialized\n";
    return(*this); // throw ...
  }

  if (isEmpty()) {
    std::cerr << "TheoraDecoder::operator>>: no packet available to create new picture\n";
    return(*this); // throw ...
  }

  /* get the first packet from the packet list */
  OggPacket packet = packetList.front();
  packetList.pop_front();

  /* insert the packet into the theora decoder */
  theora_decode_packetin(&theoraState, packet.obj());

  /* extract extra information */
  picture.setKeyFrame(theora_packet_iskeyframe(packet.obj()) == 1);
  picture.setGranulePosition(theoraState.granulepos);
  picture.setTime(theora_granule_time(&theoraState, theoraState.granulepos));

  /* finally decode the picture */
  theora_decode_YUVout(&theoraState, &picture.obj()->yuvBuffer);

  if (packetList.empty()) {
    setEmpty();
  }

  return(*this);
}

std::string TheoraDecoder::getInfoString()
{

  std::stringstream stream;

  if (!isConfigured()) {
    std::cerr << "TheoraDecoder::operator>>: Theora Stream is not initialized\n";
    return(stream.str());
  }

  stream << std::endl;
  stream << "Size         : " << theoraInfo.width << ":" << theoraInfo.height
    << " (Frame Size : " << theoraInfo.frame_width << ":" << theoraInfo.frame_height << ")\n";
  stream << "Aspect Ratio : " << theoraInfo.aspect_numerator << ":" << theoraInfo.aspect_denominator << std::endl;
  stream << "Framerate    : " << theoraInfo.fps_numerator/theoraInfo.fps_denominator << "\n";

  stream << "Colorspace   : ";
  switch (theoraInfo.colorspace) {
  case OC_CS_ITU_REC_470M:
    stream << "NTSC\n";
  break;
  case OC_CS_ITU_REC_470BG:
    stream << "PAL\n";
  break;
  default:
    stream << "unspecified\n";
  }

  stream << "Quality      : " << theoraInfo.quality << " / 61" << std::endl;

  stream << std::endl;

  stream << "Comments:\n";
  for (int i=0; i<theoraComment.comments; ++i)
    stream << theoraComment.user_comments[i] << std::endl;

  return(stream.str());
}

theora_comment& TheoraDecoder::getComment()
{
  return (theoraComment);
}

#endif
